/***
 *  Software License Agreement: BSD 3-Clause License
 *
 *  Copyright (c) 2016-2022, qbrobotics®
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 *  following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *    following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other materials provided with the distribution.
 *
 *  * Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
 *    products derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <qb_softhand_industry_driver/qb_softhand_industry_communication_handler.h>

using namespace qb_softhand_industry_communication_handler;

qbSoftHandIndustryCommunicationHandler::qbSoftHandIndustryCommunicationHandler() 
  : node_handle_(ros::NodeHandle()),
    activate_motors_(node_handle_.advertiseService("/qb_softhand_industry_communication_handler/activate_motors", &qbSoftHandIndustryCommunicationHandler::activateCallback, this)),
    deactivate_motors_(node_handle_.advertiseService("/qb_softhand_industry_communication_handler/deactivate_motors", &qbSoftHandIndustryCommunicationHandler::deactivateCallback, this)),
    set_commands_(node_handle_.advertiseService("/qb_softhand_industry_communication_handler/set_commands", &qbSoftHandIndustryCommunicationHandler::setCommandsCallback, this)),
    set_command_(node_handle_.advertiseService("/qb_softhand_industry_communication_handler/set_command", &qbSoftHandIndustryCommunicationHandler::setCommandCallback, this)),
    get_measurements_(node_handle_.advertiseService("/qb_softhand_industry_communication_handler/get_measurements", &qbSoftHandIndustryCommunicationHandler::getMeasurementsCallback, this)),
    ip_(node_handle_.param<std::string>("ip_address", "192.168.1.110")),
    serial_port_name_(node_handle_.param<std::string>("serial_port_name", "")),
    use_only_ethernet(node_handle_.param<bool>("use_only_ethernet", true)){
      if (!use_only_ethernet)
      { 
        if(serial_port_name_ == ""){
          api_handler_.reset(new qbsofthand_industry_api::qbSoftHandIndustryAPI());
        } else {
          api_handler_.reset(new qbsofthand_industry_api::qbSoftHandIndustryAPI(serial_port_name_));
        }
      } else {
        api_handler_.reset(new qbsofthand_industry_api::qbSoftHandIndustryAPI(ip_,10));
      } 
      
      while(!(api_handler_->isInitialized())){
        ros::Duration(1.0).sleep();
      }
      ROS_INFO_STREAM_NAMED("qbSoftHand Industry Communication Handler", "Connected to device with ip " << ip_ << ".");
    }

int qbSoftHandIndustryCommunicationHandler::activate(const bool &command, const int &max_repeats) {
  std::string command_prefix = command ? "" : "de";
  bool status = false;
  if(command) {
    status = api_handler_->sendRawCommand("MO=1",max_repeats); // activation command
  } else {
    status = api_handler_->sendRawCommand("MO=0",max_repeats); // deactivation command
  }
  ROS_INFO_STREAM_NAMED("qbSoftHand Industry Communication Handler", "device [" << ip_ << "] motors have been " << command_prefix << "activated!");
  return status;
}

bool qbSoftHandIndustryCommunicationHandler::activateCallback(qb_softhand_industry_srvs::TriggerRequest &request, qb_softhand_industry_srvs::TriggerResponse &response) {
  ROS_ERROR_STREAM_COND_NAMED(request.max_repeats < 0, "communication_handler", "Device [" << ip_ << "] has request service with non-valid 'max_request' [" << request.max_repeats << "].");
  activate(true, request.max_repeats);
  response.message = "Device [" + ip_ + "] activation.";
  return true;
}

bool qbSoftHandIndustryCommunicationHandler::deactivateCallback(qb_softhand_industry_srvs::TriggerRequest &request, qb_softhand_industry_srvs::TriggerResponse &response) {
  ROS_ERROR_STREAM_COND_NAMED(request.max_repeats < 0, "communication_handler", "Device [" << ip_ << "] has request service with non-valid 'max_request' [" << request.max_repeats << "].");
  activate(false, request.max_repeats);
  response.message = "Device [" + ip_ + "] deactivation.";
  return true;
}

bool qbSoftHandIndustryCommunicationHandler::getMeasurementsCallback(qb_softhand_industry_srvs::GetMeasurementsRequest &request, qb_softhand_industry_srvs::GetMeasurementsResponse &response) {
  ROS_ERROR_STREAM_COND_NAMED(request.max_repeats < 0, "communication_handler", "Device [" << ip_ << "] has request service with non-valid 'max_request' [" << request.max_repeats << "].");
  if(request.get_position) {
    response.position = QBSOFTHAND_INDUSTRY_POS_MIN + api_handler_->getPosition() * QBSOFTHAND_INDUSTRY_POS_MAX * 0.01;
  }
  if(request.get_current) {
    response.current = QBSOFTHAND_INDUSTRY_CURR_MIN + api_handler_->getCurrent() * QBSOFTHAND_INDUSTRY_CURR_MAX * 0.01; 
  }
  if(request.get_velocity) {
    response.velocity = QBSOFTHAND_INDUSTRY_VEL_MIN +api_handler_->getVelocity() * QBSOFTHAND_INDUSTRY_VEL_MAX * 0.01;
  }

  response.stamp = ros::Time::now();
  response.success = true;
  return true;
}

bool qbSoftHandIndustryCommunicationHandler::isValidCurrent(const int &current){
  return current >= QBSOFTHAND_INDUSTRY_CURR_MIN && current <= QBSOFTHAND_INDUSTRY_CURR_MAX;
}

bool qbSoftHandIndustryCommunicationHandler::isValidPosition(const int &position){
  return position >= QBSOFTHAND_INDUSTRY_POS_MIN && position <= QBSOFTHAND_INDUSTRY_POS_MAX;
}

bool qbSoftHandIndustryCommunicationHandler::isValidVelocity(const int &velocity){
  return velocity >= QBSOFTHAND_INDUSTRY_VEL_MIN && velocity <= QBSOFTHAND_INDUSTRY_VEL_MAX;
}

int qbSoftHandIndustryCommunicationHandler::setCommand(const int &command, const int &max_repeats) {
  if(api_handler_->sendRawCommand("SR", max_repeats) != 0) {
    ROS_WARN_STREAM_NAMED("qbSoftHand Industry Communication Handler", "Device not responding");
    return QBSOFTHAND_INDUSTRY_ERROR_NOTRESPONDING;
  }
  if(!isValidPosition(command)){
    return QBSOFTHAND_INDUSTRY_ERROR_BADPARAMS;
  }

  api_handler_->sendRawCommand("UI[1]=" + std::to_string(command), max_repeats);
  return 0;
}

int qbSoftHandIndustryCommunicationHandler::setCommands(const int &position, const int &velocity, const int &current, const int &max_repeats) {
  if(api_handler_->sendRawCommand("SR", max_repeats) != 0) {
    ROS_WARN_STREAM("Device not responding");
    return QBSOFTHAND_INDUSTRY_ERROR_NOTRESPONDING;
  }

  if(!isValidPosition(position) || !isValidVelocity(velocity) || !isValidCurrent(current)){
    return QBSOFTHAND_INDUSTRY_ERROR_BADPARAMS;
  }

  api_handler_->sendRawCommand("UI[2]=" + std::to_string(velocity), max_repeats);
  api_handler_->sendRawCommand("UI[3]=" + std::to_string(current), max_repeats);
  api_handler_->sendRawCommand("UI[1]=" + std::to_string(position), max_repeats);
  return 0;
}

bool qbSoftHandIndustryCommunicationHandler::setCommandCallback(qb_softhand_industry_srvs::SetCommandRequest &request, qb_softhand_industry_srvs::SetCommandResponse &response) {
  ROS_ERROR_STREAM_COND_NAMED(request.max_repeats < 0, "communication_handler", "Device [" << ip_ << "] has request service with non-valid 'max_request' [" << request.max_repeats << "].");

  if (request.set_commands) {
    response.success = (setCommand(request.position_command, request.max_repeats) == 0);
  }
  return true;
}

bool qbSoftHandIndustryCommunicationHandler::setCommandsCallback(qb_softhand_industry_srvs::SetCommandsRequest &request, qb_softhand_industry_srvs::SetCommandsResponse &response) {
  ROS_ERROR_STREAM_COND_NAMED(request.max_repeats < 0, "communication_handler", "Device [" << ip_ << "] has request service with non-valid 'max_request' [" << request.max_repeats << "].");
  if (request.set_commands) {
    response.success = (setCommands(request.position_command, request.velocity_command, request.current_command, request.max_repeats) == 0);
  }
  return true;
}
